//! Control program for the ActivityPub federated social media server.
#![feature(iterator_try_collect)]

use puppy::{
    actor::Actor,
    config::Config,
    data::{FollowRequest, Object, Profile},
    post::Author,
    Context,
};

#[tokio::main]
async fn main() -> puppy::Result<()> {
    // puppy::store::Store::nuke(".state")?;
    let config = Config {
        ap_domain: "test.piss-on.me".to_string(),
        wf_domain: "test.piss-on.me".to_string(),
        state_dir: ".state".to_string(),
        port: 1312,
    };
    let cx = Context::load(config)?;
    let riley = get_or_create_actor(&cx, "riley")?;
    cx.run(|tx| {
        println!("\nRiley's following:");
        for FollowRequest { id, origin, .. } in
            riley.pending_requests(&tx).try_collect::<Vec<_>>()?
        {
            let Profile { account_name, .. } = tx.get_mixin(origin)?.unwrap();
            let Object { id, .. } = tx.get_mixin(id)?.unwrap();
            println!("- @{account_name} ({origin}) (request url = {id})");
        }
        Ok(())
    })?;
    // let post = puppy::post::create_post(&cx, riley.key, "i like boys")?;
    // puppy::post::federate_post(&cx, post).await

    // let linen = get_or_create_actor(&cx, "linen")?;
    // if true {
    //     println!("creating posts");
    //     puppy::post::create_post(&cx, riley.key, "@linen <3")?;
    //     puppy::post::create_post(&cx, linen.key, "@riley <3")?;
    // }

    // if true {
    //     println!("making riley follow linen");
    //     cx.run(|tx| {
    //         if !riley.follows(&tx, &linen)? {
    //             println!("follow relation does not exist yet");
    //             if let Some(req) = linen
    //                 .pending_requests(&tx)
    //                 .find_ok(|r| r.origin == riley.key)?
    //             {
    //                 println!("accepting the pending follow request");
    //                 linen.do_accept_request(&cx, req)
    //             } else {
    //                 println!("no pending follow request; creating");
    //                 riley.do_follow_request(&cx, &linen).map(|_| ())
    //             }
    //         } else {
    //             println!("riley already follows linen");
    //             Ok(())
    //         }
    //     })?;
    // }

    println!("\nPosts on the instance:");
    for post in puppy::post::fetch_timeline(cx.store(), .., None)?.posts() {
        let Author { ref handle, .. } = post.author;
        let content = post.content.content.as_ref().unwrap();
        println!("- {:?} by {handle}:\n{content}", post.id)
    }
    Ok(())

    // cx.run(|tx| {
    //     println!("\nLinen's followers:");
    //     for id in linen.followers(&tx).try_collect::<Vec<_>>()? {
    //         let Profile { account_name, .. } = db.get_mixin(id)?.unwrap();
    //         println!("- @{account_name} ({id})");
    //     }

    //     println!("\nRiley's following:");
    //     for id in riley.following(&tx).try_collect::<Vec<_>>()? {
    //         let Profile { account_name, .. } = db.get_mixin(id)?.unwrap();
    //         println!("- @{account_name} ({id})");
    //     }

    //     if false {
    //         println!("Biting riley");
    //         linen.do_bite(&cx, &riley)?;
    //         for Bite { id, biter, .. } in riley.bites_suffered(&tx).try_collect::<Vec<_>>()? {
    //             let Profile { account_name, .. } = db.get_mixin(biter)?.unwrap();
    //             println!("riley was bitten by @{account_name} at {}", id.timestamp());
    //         }
    //     }
    //     Ok(())
    // })
}

fn get_or_create_actor(cx: &Context, username: &str) -> puppy::Result<Actor> {
    let user = cx.run(|tx| Actor::by_username(tx, username))?;
    match user {
        Some(key) => {
            println!("found '{username}' ({key:?})");
            Ok(key)
        }
        None => {
            println!("'{username}' doesn't exist yet, creating");
            let r = puppy::actor::create_local(cx, username);
            if let Ok(ref key) = r {
                println!("created '{username}' with key {key:?}");
            }
            r
        }
    }
}
