//! ActivityPub handlers.

use http_body_util::Full;
use hyper::body::Bytes;
use puppy::{
    actor::{get_signing_key, Actor},
    fetch::object::Activity,
    get_local_ap_object, Context, Error, Key,
};
use serde_json::Value;
use tracing::{info, instrument};

use crate::sig::{Signer, Verifier};
use super::{error::Message, Response};

/// Proxy a request through the instance.
pub async fn proxy(cx: &Context, params: &[(&str, &str)]) -> Result<Response, Message> {
    // Extract our query parameters.
    let Some(user) = params.iter().find_map(|(k, v)| (*k == "user").then_some(v)) else {
        fuck!(400: "expected `user` query param");
    };
    let Some(url) = params.iter().find_map(|(k, v)| (*k == "url").then_some(v)) else {
        fuck!(400: "expected `url` query param");
    };

    // Look up the actor's key in the store (which is accessible through the puppy context).
    let Ok(signing_key) = cx.run(|tx| {
        let actor = Actor::by_username(&tx, user)?.unwrap();
        get_signing_key(tx, actor).map_err(Error::from)
    }) else {
        fuck!(500: "failed to get signing key");
    };

    eprintln!("proxy: params: {params:?}");
    // Proxy the request through our fetcher.
    let resp = puppy::fetch::forward(&signing_key, url).await.unwrap();
    eprintln!("proxy: status = {}", resp.status());

    // Convert the http-types request to a hyper request.
    Ok(resp.map(Bytes::from).map(Full::new).into())
}

pub async fn outbox(cx: &Context, params: &[(&str, &str)]) -> Result<Response, Message> {
    // Extract our query parameters.
    let Some(user) = params.iter().find_map(|(k, v)| (*k == "user").then_some(v)) else {
        fuck!(400: "expected `user` query param");
    };
    let Some(content) = params
        .iter()
        .find_map(|(k, v)| (*k == "content").then_some(v))
    else {
        fuck!(400: "expected `url` query param");
    };

    let Ok(Some(actor)) = cx.run(|tx| Actor::by_username(&tx, user)) else {
        fuck!(500: "failed actor by name {user}");
    };

    let post = puppy::post::create_local_post(&cx, actor.key, content.to_string()).unwrap();
    puppy::post::federate_post(&cx, post).await.unwrap();
    Ok(respond! {
        code: 200
    })
}

/// Handle POSTs to actor inboxes. Requires request signature.
#[instrument(skip_all)]
pub async fn inbox(
    cx: &Context,
    actor_id: &str,
    sig: Signer,
    body: &[u8],
) -> Result<Response, Message> {
    let receiver = actor_id.parse::<Key>().unwrap();
    let json: Value = serde_json::from_slice(body).unwrap();
    let id = json["id"].to_string();

    info! {
        inbox = receiver.to_string(),
        signature = sig.ap_id,
        "processing object '{id}'",
    };

    match Activity::from_json(json) {
        Ok(activity) => {
            puppy::ingest(&cx, receiver, &activity).await.unwrap();
            match puppy::interpret(&cx, activity) {
                Ok(_) => Ok(respond!(code: 202)),
                Err(err) => fuck!(400: "error interpreting activity: {err}"),
            }
        }
        Err(err) => fuck!(400: "invalid payload: {err}"),
    }
}

/// Serve an ActivityPub object as json-ld.
pub fn serve_object(cx: &Context, object_ulid: &str) -> Result<Response, Message> {
    let Ok(parsed) = object_ulid.parse::<Key>() else {
        fuck!(400: "improperly formatted ulid");
    };

    let result = cx.run(|tx| get_local_ap_object(&tx, parsed));
    let Ok(object) = result else {
        fuck!(404: "object does not exist");
    };

    Ok(respond! {
        kind: "application/activity+json",
        body: Some(&object.to_json_ld())
    })
}

/// Serve the special actor used for signing requests.
pub fn serve_verifier_actor(verifier: &Verifier) -> Response {
    respond! {
        kind: "application/activity+json",
        body: Some(&verifier.to_json_ld())
    }
}
