//! WebFinger endpoints and related stuff.

use puppy::{
    data::{Id, Username},
    Context,
};
use serde_json::{json, Value};
use derive_more::Display;

use super::{error::Message, Response};

const WF_CONTENT_TYPE: (&str, &str) = ("content-type", "application/jrd+json");

/// Respond to a webfinger request.
pub fn resolve(cx: &Context, params: &[(&str, &str)]) -> Result<Response, Message> {
    match params.iter().find_map(get_handle) {
        Some(handle) if cx.config().wf_domain == handle.instance => {
            let username = Username(handle.username.to_string());
            let Ok(Some(user)) = cx.store().lookup(username) else {
                fuck!(404: "no user {}@{} exists", handle.username, handle.instance);
            };
            let Ok(Some(Id(id))) = cx.store().get_alias(user) else {
                fuck!(500: "internal error");
            };
            let jrd = make_jrd(handle, &id);
            Ok(respond! {
                body: Some(&jrd),
                kind: "application/jrd+json"
            })
        }
        Some(_) | None => fuck!(400: "missing/invalid resource param"),
    }
}

#[derive(Clone, Copy, Display)]
#[display(fmt = "@{username}@{instance}")]
pub struct Handle<'x> {
    username: &'x str,
    instance: &'x str,
}

/// Parse the `resource` parameter into a [`Handle`].
fn get_handle<'x>((k, v): &'x (&str, &str)) -> Option<Handle<'x>> {
    // We're looking for the `resource` query parameter.
    if *k == "resource" {
        // This prefix needs to exist according to spec.
        let (username, instance) = v
            .strip_prefix("acct:")?
            // Some implementations may prefix with `@`. its ok if it's there and its also ok
            // if its not there, so we use `trim_start_matches` instead of `strip_prefix`.
            .trim_start_matches('@')
            // Split on the middle `@` symbol, which separates the username and instance bits
            .split_once('@')?;
        Some(Handle { username, instance })
    } else {
        None
    }
}

/// Construct a "JSON resource descriptor".
fn make_jrd(handle: Handle<'_>, id: &str) -> Value {
    json!({
      "subject": format!("acct:{}@{}", handle.username, handle.instance),
      "links": [
        {
          "rel": "self",
          "type": "application/activity+json",
          "href": id
        },
      ]
    })
}
