use proc_macro::TokenStream;

mod arrow;

#[proc_macro_derive(Arrow, attributes(origin, target, identity))]
pub fn arrow(item: TokenStream) -> TokenStream {
    arrow::arrow(item)
}

#[proc_macro_derive(Alias)]
pub fn alias(item: TokenStream) -> TokenStream {
    let input = syn::parse_macro_input!(item as syn::DeriveInput);
    let syn::Data::Struct(structure) = input.data else {
        panic!("Only structs are supported as aliases")
    };
    match structure.fields {
        syn::Fields::Unnamed(f) if f.unnamed.len() == 1 => {
            let first = f.unnamed.first().unwrap();
            make_alias_impl(&input.ident, first)
        }
        _ => panic!("Only string newtype structs are allowed as aliases"),
    }
}

fn make_alias_impl(name: &syn::Ident, field: &syn::Field) -> TokenStream {
    let typ = &field.ty;
    let prefix = heck::AsKebabCase(name.to_string());
    let keyspace = format!("{prefix}/keyspace");
    let reversed = format!("{prefix}/reversed");
    let spec = quote::quote! {
        impl store::types::DataType for #name {
            type Type = store::types::AliasSpec;
            const SPEC: Self::Type = store::types::AliasSpec {
                keyspace: store::types::Keyspace(#keyspace),
                reversed: store::types::Keyspace(#reversed),
            };
        }
    };

    TokenStream::from(quote::quote! {
        #spec
        impl store::Alias for #name {}
        impl AsRef<str> for #name {
            fn as_ref(&self) -> &str { self.0.as_ref() }
        }
        impl From<#typ> for #name {
            fn from(v: #typ) -> #name { #name(v) }
        }
    })
}

#[proc_macro_derive(Mixin, attributes(index))]
pub fn mixin(item: TokenStream) -> TokenStream {
    let input = syn::parse_macro_input!(item as syn::DeriveInput);

    let name = input.ident;
    let prefix = heck::AsKebabCase(name.to_string());
    let keyspace = format!("{prefix}/main");

    let spec = quote::quote! {
        impl store::types::DataType for #name {
            type Type = store::types::MixinSpec;
            const SPEC: Self::Type = store::types::MixinSpec {
                keyspace: store::types::Keyspace(#keyspace),
            };
        }
    };

    TokenStream::from(quote::quote! {
        #spec
        impl store::Mixin for #name {}
    })
}
